/***************************************************************************
 *   Copyright (C) 2010 by Vladimir Jary                                   *
 *   <Vladimir.Jary@cern.ch>                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "speexporter.h"

#include<QtCore/QFile>
#include<QtCore/QString>
#include<QtCore/QIODevice>
#include<QtCore/QDataStream>
#include<QtCore/QTextStream>
#include<QtCore/QStringList>

#include "iostream"


/*
  Initialization of input and output streams and parses
  header of input SPE file
*/
SpeExporter::SpeExporter(QString input)
{
    inFile = new QFile();
    outFile = new QFile();
    inputName = input;
    inFile->setFileName(input);
    inFile->open(QIODevice::ReadOnly);
    in = new QDataStream(inFile);
    out = new QTextStream;
    newLine = "\n";
    setNewLine(UNIX);
    parseHeader();
    separator=';';
}

/*
  Closes and deletes specified file
*/
void SpeExporter::deleteFile(QFile *file)
{
    if (file){
        if (file->isOpen()) file->close();
        delete file;
        file = 0;
    }
}

/*
   Clean-up, destroys files
 */
SpeExporter::~SpeExporter()
{
    deleteFile(inFile);
    deleteFile(outFile);
}

/*
  Sets the newline sequence according to desired platform:
  DOS/Windows: CR, LF
  Unix: LF
  Macintosh: CR
  see http://en.wikipedia.org/wiki/Newline
*/
void SpeExporter::setNewLine(Platform platform)
{
    if(platform == DOS)
        newLine = "\r\n"; //DOS
    else if(platform == MAC)
        newLine = "\r"; //Macintosh
    else newLine = "\n"; //Unix
}

/*
  Prints relevant information from SPE file header:
  frame dimension, number of frames, and data format
*/
void SpeExporter::printHeader()
{
    std::cout << "-----------------------" << std::endl;
    std::cout << "Input: " << qPrintable(inputName) << std::endl;
    std::cout << "x: " << x << std::endl;
    std::cout << "y: " << y << std::endl;
    std::cout << "Size: " << size << " " << std::endl;
    std::cout << "Number of frames: " << frames << std::endl;
    std::cout << "Data format: " << format << std::endl;
    std::cout << "-----------------------" << std::endl;
}

/*
  Processes header of opened SPE file
  x ... stored at position 42
  y ... stored at position 656
*/
void SpeExporter::parseHeader()
{

    in->skipRawData(34);
    frames = readUInt16(in);

    if(frames == 65535) frames = 1;

    in->skipRawData(6);
    x = readUInt16(in);


    in->skipRawData(64);
    format = readUInt16(in);

    //do we use supported format?
    if((format != 3) && (format != 1)){
        std::cerr << "Warning: input file seems to be in unsupported format!" << std::endl;
    }

    in->skipRawData(546);
    y = readUInt16(in);
    size = x*y;

    printHeader();

    //skip rest of header
    in->skipRawData(3442);

}

/*
  Reads unsigned 16bit integer from input data stream
 */
quint16 SpeExporter::readUInt16(QDataStream* instream)
{
    qint8 byte1, byte2;
    qint32 b1, b2;
    qint16 word;
    instream->setByteOrder(QDataStream::LittleEndian);
    instream->operator >>(byte1);
    instream->operator >>(byte2);

    b1 = byte1;
    b2 = byte2;
    if (b1 < 0) b1+=256;
    if (b2 < 0) b2+=256;
    word = (b2 << 8) | b1;

    return word;
}

/*
  Reads signed 32bit integer from input data stream
 */

qint32 SpeExporter::readInt32(QDataStream *instream)
{
    qint32 intensity;
    instream->operator >> (intensity);
   // std::cout << intensity << std::endl;
    return intensity;
}


/*
  Prepares input and output streams
 */
void SpeExporter::openStreams()
{
    in->setDevice(inFile);
    outFile->setFileName(outputName);
    outFile->open(QIODevice::WriteOnly);
    out->setDevice(outFile);
}

/*
  Closes opened files
 */
void SpeExporter::closeStreams()
{
    inFile->close();
    outFile->close();
}


/*
  Allocates dynamic 2D array with given rows and columns
*/
qint32** SpeExporter::Allocate2DArray(quint32 rows, quint32 cols)
{
    qint32** buff;
    buff = new qint32*[rows];
    for(quint32 i = 0; i < rows; i++)
    {
        buff[i] = new qint32[cols];
    }
    return buff;
}

/*
  Releases memory occupied by 2D array buff
*/
void SpeExporter::Free2DArray(qint32**buff)
{
    delete [] *buff;
    delete [] buff;
}


/*
  Loads data from input stream in into 2D array buff
*/
void SpeExporter::loadData(qint32**buff, QDataStream *in)
{
    //quint16 intensity;
    qint32 intensity;

    for(quint16 i = 0; i < frames; i++)
    {
        for(quint32 j = 0; j < size; j++){
            if (format == 3){
                intensity = readUInt16(in);
            } else
            {
                intensity = readInt32(in);
                //std::cout << intensity << std::endl;
             }
            buff[i][j] = intensity;
          /*  if(in->atEnd())
                std::cout << "End of file reached" << std::endl;*/

        }
    }

}

/*
  Converts SPE file into plain-text format
 */
void SpeExporter::export2txt()
{
    constructOutputName(".txt");
    openStreams();

    //quint16 **buffer = Allocate2DArray(frames, size);
    qint32 **buffer = Allocate2DArray(frames, size);


    loadData(buffer, in);

    for(quint32 i = 0; i < size; i++)
    {
        for(quint32 j = 0; j < frames; j++){
            out->operator << (buffer[j][i]);
            if(j+1 == frames) out->operator <<(newLine);
                else out->operator <<(' ');
        }
    }

    Free2DArray(buffer);
    closeStreams();
}


/*
  Converts SPE file into plain-text format, each scan (column)
  is stored into separate ASCII file (*.asc)
*/
void SpeExporter::export2asc()
{
    qint32 intensity;
    quint16 scan = 0;
    QString suffix;


    for(quint32 j = 0; j < size*frames; j++){
        if(j % size == 0)
        {
            outFile->close();
            scan+=1;
            suffix.setNum(scan);
            if(scan < 10)
                suffix.prepend("_000");
            else if(scan < 100)
                suffix.prepend("_00");
            else if(scan < 1000)
                suffix.prepend("_0");
            else suffix.prepend("_");

            suffix.append(".asc");
            constructOutputName(suffix);


            //std::cout << scan <<  " " << qPrintable(inputName) << " " << qPrintable(outputName) << std::endl;
            openStreams();
        }

        if (format == 3){
            intensity = readUInt16(in);
        } else
        {
            intensity = readInt32(in);
            //std::cout << intensity << std::endl;
         }
        out->operator <<((float)intensity) << (newLine);

    }
    closeStreams();
}


/*
  Converts input SPE file into comma-separated value
  spreadsheet
 */
void SpeExporter::export2csv()
{
    constructOutputName(".csv");
    openStreams();

    qint32 **buffer = Allocate2DArray(frames, size);
    loadData(buffer, in);

    for(quint32 i = 0; i < size; i++)
    {
        out->operator <<(i+1) << separator;
        for(quint32 j = 0; j < frames; j++){
            out->operator << (buffer[j][i]);
            if(j+1 == frames) out->operator <<(newLine);
                else out->operator <<(',');
        }
    }

    Free2DArray(buffer);
    closeStreams();
}


/*
  Converts input SPE file into portable graymap
 */
void SpeExporter::export2pgm()
{
    constructOutputName(".pgm");
    openStreams();
    qint32 intensity;
    //max pixel intensity
    qint32 max = 0;
    QString line = "";
    //image data
    QStringList lines;

    for(quint32 j = 1; j <= size*frames; j++){
        if (format == 3){
                       intensity = readUInt16(in);
                   } else
                   {
                       intensity = readInt32(in);

                    }
        if (intensity > max)
            max = intensity;
        line.append(QString("%1").arg(intensity));
        if(line.length() < 65)
            line.append(' ');
        else
        {
            line.append(newLine);
            lines.push_back(line);
            line = "";
        }

    }
    lines.push_back(line);

    //write header
    out->operator <<("P2") << (newLine);
    out->operator << (x) << (' ') << (y*frames) << (newLine);
    out->operator <<(max) << (newLine);

    //write data
    QStringListIterator it(lines);
    while(it.hasNext())
    {
        out->operator <<(it.next());
    }

    lines.clear();
    closeStreams();
}

/*
  Generates name for output file - same as input, but in lowercase
  and with changed extension (SPE -> csv/pgm/txt)
 */
void SpeExporter::constructOutputName(QString extension)
{
  //outputName = inputName.toLower();
    QString tmp = inputName;
    outputName = inputName.replace(".spe", extension, Qt::CaseInsensitive);
    inputName = tmp;
}

/*
  Returns width in pixels
 */
inline quint16 SpeExporter::getX()
{
    return x;
}

/*
  Returns height in pixels
 */
inline quint16 SpeExporter::getY()
{
    return y;
}

/*
  Returns number of frames stored in SPE file
 */
inline quint16 SpeExporter::getNumberOfFrames()
{
    return frames;
}
